/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <gstdisksrc.h>


GstElementDetails gst_disksrc_details = {
  "Disk Source",
  "Source/File",
  "Synchronous read from a file",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* DiskSrc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_LOCATION,
  ARG_BYTESPERREAD
};


static void gst_disksrc_class_init(GstDiskSrcClass *klass);
static void gst_disksrc_init(GstDiskSrc *disksrc);
static void gst_disksrc_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_disksrc_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstSrcClass *parent_class = NULL;
static guint gst_disksrc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_disksrc_get_type(void) {
  static GtkType disksrc_type = 0;

  if (!disksrc_type) {
    static const GtkTypeInfo disksrc_info = {
      "GstDiskSrc",
      sizeof(GstDiskSrc),
      sizeof(GstDiskSrcClass),
      (GtkClassInitFunc)gst_disksrc_class_init,
      (GtkObjectInitFunc)gst_disksrc_init,
      (GtkArgSetFunc)gst_disksrc_set_arg,
      (GtkArgGetFunc)gst_disksrc_get_arg,
      (GtkClassInitFunc)NULL,
    };
    disksrc_type = gtk_type_unique(GST_TYPE_SRC,&disksrc_info);
  }
  return disksrc_type;
}

static void
gst_disksrc_class_init(GstDiskSrcClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstSrcClass *gstsrc_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstsrc_class = (GstSrcClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_SRC);

  gtk_object_add_arg_type("GstDiskSrc::location", GTK_TYPE_STRING,
                          GTK_ARG_READWRITE, ARG_LOCATION);
  gtk_object_add_arg_type("GstDiskSrc::bytesperread", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_BYTESPERREAD);

  gtkobject_class->set_arg = gst_disksrc_set_arg;
  gtkobject_class->get_arg = gst_disksrc_get_arg;

  gstsrc_class->push = gst_disksrc_push;
  /* we nominally can't (won't) do async */
  gstsrc_class->push_region = NULL;
}

static void gst_disksrc_init(GstDiskSrc *disksrc) {
  disksrc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(disksrc),disksrc->srcpad);

  disksrc->filename = NULL;
  disksrc->fd = 0;
  disksrc->curoffset = 0;
  disksrc->seq = 0;
  disksrc->bytes_per_read = 4096;
}

GstElement *gst_disksrc_new(gchar *name) {
  GstElement *disksrc = GST_ELEMENT(gtk_type_new(GST_TYPE_DISKSRC));
  gst_object_set_name(GST_OBJECT(disksrc),name);
  return disksrc;
}

GstElement *gst_disksrc_new_with_location(gchar *name,gchar *location) {
  GstElement *disksrc = gst_disksrc_new(name);
  g_print("have disksrc\n");
  gtk_object_set(GTK_OBJECT(disksrc),"location",location,NULL);
  g_print("have set location\n");
  return disksrc;
}

void gst_disksrc_push(GstSrc *src) {
  GstDiskSrc *disksrc;
  GstBuffer *buf;
  glong readbytes;

  g_return_if_fail(src != NULL);
  g_return_if_fail(GST_IS_DISKSRC(src));
  disksrc = GST_DISKSRC(src);

  buf = gst_buffer_new();
  g_return_if_fail(buf);
  buf->data = (gpointer)malloc(disksrc->bytes_per_read);
  gst_trace_add_entry(NULL,0,buf,"disksrc: filling buffer");
  readbytes = read(disksrc->fd,buf->data,disksrc->bytes_per_read);
  if (readbytes == 0) {
    gst_src_signal_eof(GST_SRC(disksrc));
    return;
  }

  buf->srcoffset = disksrc->curoffset;
  buf->datasize = readbytes;
  disksrc->curoffset += readbytes;

//  g_print("gst_disksrc_push(): pushing %d byte buffer from '%s'\n",
//          buf->datasize,gst_object_get_name(GST_OBJECT(disksrc)));

  gst_pad_push(disksrc->srcpad,buf);
}

void gst_disksrc_set_filename(GstDiskSrc *disksrc,gchar *filename) {
  g_return_if_fail(disksrc != NULL);
  g_return_if_fail(GST_IS_DISKSRC(disksrc));
  g_return_if_fail(disksrc->filename == NULL);
  g_return_if_fail(filename != NULL);

  disksrc->filename = g_strdup(filename);
  disksrc->fd = open(disksrc->filename,O_RDONLY);
  if (disksrc->fd <= 0)
    perror("opening file");
/*  disksrc->size = lseek(disksrc->fd,0,SEEK_END);
  lseek(disksrc->fd,0,SEEK_SET);*/
}

gchar *gst_disksrc_get_filename(GstDiskSrc *disksrc) {
  g_return_if_fail(disksrc != NULL);
  g_return_if_fail(GST_IS_DISKSRC(disksrc));

  return disksrc->filename;
}


static void gst_disksrc_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstDiskSrc *disksrc;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_DISKSRC(object));
  disksrc = GST_DISKSRC(object);

  switch(id) {
    case ARG_LOCATION:
      g_print("gst_disksrc_set_arg: setting location to %s\n",
              GTK_VALUE_STRING(*arg));
      gst_disksrc_set_filename(disksrc,GTK_VALUE_STRING(*arg));
      break;
    case ARG_BYTESPERREAD:
      disksrc->bytes_per_read = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void gst_disksrc_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstDiskSrc *disksrc;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_DISKSRC(object));
  disksrc = GST_DISKSRC(object);

  switch (id) {
    case ARG_LOCATION:
      GTK_VALUE_STRING(*arg) = gst_disksrc_get_filename(disksrc);
      break;
    case ARG_BYTESPERREAD:
      GTK_VALUE_INT(*arg) = disksrc->bytes_per_read;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}
