/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "gstmpg123.h"


/* elementfactory information */
GstElementDetails gst_mpg123_details = {
  "mpg123 mp3 decoder",
  "Filter/Decoder/Audio",
  "Uses modified mpg123 code to decode mp3 streams",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* Mpg123 signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_mpg123_class_init(GstMpg123Class *klass);
static void gst_mpg123_init(GstMpg123 *mpg123);


static GstFilterClass *parent_class = NULL;
static guint gst_mpg123_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_mpg123_get_type(void) {
  static GtkType mpg123_type = 0;

  if (!mpg123_type) {
    static const GtkTypeInfo mpg123_info = {
      "GstMpg123",
      sizeof(GstMpg123),
      sizeof(GstMpg123Class),
      (GtkClassInitFunc)gst_mpg123_class_init,
      (GtkObjectInitFunc)gst_mpg123_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    mpg123_type = gtk_type_unique(GST_TYPE_FILTER,&mpg123_info);
  }
  return mpg123_type;
}

static void
gst_mpg123_class_init(GstMpg123Class *klass) {
  GstFilterClass *gstfilter_class;

  gstfilter_class = (GstFilterClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_FILTER);
}

static void gst_mpg123_init(GstMpg123 *mpg123) {
  /* create the sink and src pads */
  mpg123->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(mpg123),mpg123->sinkpad);
  gst_pad_set_chain_function(mpg123->sinkpad,gst_mpg123_chain);
  mpg123->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(mpg123),mpg123->srcpad);

  /* initialize the mpg123 decoder state */
  mpg123_init_layer3(&mpg123->decoder,32);
  mpg123->frame = mpg123_init_frame(&mpg123->decoder);
  mpg123->prevheader = 0;
  mpg123->prevbuf = NULL;
  mpg123->lastframe = 0;
  mpg123->framebuf = NULL;
  mpg123->framebufsize = 0;
}

GstElement *gst_mpg123_new(gchar *name) {
  GstElement *mpg123 = GST_ELEMENT(gtk_type_new(GST_TYPE_MPG123));
  gst_object_set_name(GST_OBJECT(mpg123),name);
  return mpg123;
}

void gst_mpg123_chain(GstPad *pad,GstBuffer *buf) {
  GstMpg123 *mpg123;
  /* buffer info */
  guchar *bufdata, *data;
  gulong size;
  /* reconstruction pointers */
  guchar *framedata;
  /* per-buffer state */
  gint frames = 0;
  /* pre-frame state */
  gulong header;
  guchar mpegver;
  guchar layer;
  guchar bitrateindex;
  guchar samplerateindex;
  guint bitrate;
  guint samplerate;
  gdouble tpf;
  guchar outbuf[1152*4];
//  gulong offset;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  g_return_if_fail(GST_IS_BUFFER(buf));

  mpg123 = GST_MPG123(pad->parent);

  bufdata = (guchar *)buf->data;
  data = bufdata;
  size = buf->datasize;
//  g_print("gst_mpg123_chain: got buffer of %d bytes in '%s'\n",size,
//          gst_object_get_name(GST_OBJECT(mpg123)));

/*
  // first deal with an underflowed frame in previous buffer
  // the mpg123->frame structure is valid for previous frame
  if (mpg123->prevbuf != NULL) {
    g_print("have partial buffer left over\n");
    // first verify we have a buffer of sufficient size to construct to
    if (mpg123->framebuf == NULL)
      mpg123->framebuf = (guchar *)malloc(mpg123->frame->bpf);
    else if (mpg123->framebufsize < mpg123->frame->bpf) {
      free(mpg123->framebuf);
      mpg123->framebuf = (guchar *)malloc(mpg123->frame->bpf);
    }

    // now copy the first part of the frame over
    memcpy(mpg123->framebuf,mpg123->prevbuf->data+mpg123->lastframe,
           mpg123->lastframebytes);
    // and free the buffer
    gst_buffer_unref(mpg123->prevbuf);

    // now put the second part in
    memcpy(mpg123->framebuf + mpg123->lastframebytes,data,
           mpg123->frame->bpf - mpg123->lastframebytes);

    data += (mpg123->frame->bpf - mpg123->lastframebytes);

    framedata = mpg123->framebuf;

//    g_print("we have a prev frame: MPEG v%s, layer %d, %d Kbps, %d bytes\n",
//            (mpg123->frame->mpegver == MPG_VER_1) ? "1.0" :
//            (mpg123->frame->mpegver == MPG_VER_2) ? "2.0" :
//            (mpg123->frame->mpegver == MPG_VER_25) ? "2.5" : "UNK",
//            (mpg123->frame->layer == MPG_LAYER_1) ? 1 :
//            (mpg123->frame->layer == MPG_LAYER_2) ? 2 :
//            (mpg123->frame->layer == MPG_LAYER_3) ? 3 : 0,
//            mpg123->frame->bitrate,mpg123->frame->bpf);

//    fr->decoder = &mpg123->decoder;
//    mpg123_decode_frame(&mpg123->frame,mpg123->framebuf,outbuf);
    frames++;
  }
*/

  /* loop until we're out of buffer */
  while ((data - bufdata) < size) {
//    offset = 0;
    /* first seach for the beginning the possible beginning of a frame */
    while (*data != 0xff) {
      data++;
      mpg123->skipped_bytes++;
//      offset++;
      /* if we're at the end of the buffer with no frames, squawk */
      if ((data - bufdata) == size) {
        mpg123->skipped_buffers++;
        gst_buffer_unref(buf);
        /* FIXME: fire off a signal here */
//        g_print("sorry, couldn't find a byte of sync\n");
        return;
      }
    }
//    if (offset)
//      g_print("skipped %d bytes to find potential header\n",offset);

    /* construct the frame header */
    header = GULONG_TO_BE(*((gulong *)data));
    /* now check the potential frame header */
    if (!mpg123_head_check(header))
      break;

    /* at this point we should have a valid MPEG audio frame header */

    /* if we haven't already decoded this header... */
    if (header != mpg123->prevheader) {
//      g_print("decoding header 0x%08x\n",header);
      /* decode the header */
      mpg123_decode_header(mpg123->frame,data);
    }
    mpg123->prevheader = header;

    /* deal with the case where we don't have enough data this buffer */
    if ((size - (data-bufdata)) < mpg123->frame->bpf) {
//      g_print("frame underflow, need next buffer to complete frame\n");
      mpg123->prevbuf = buf;
      mpg123->lastframe = data - bufdata;
      mpg123->lastframebytes = size - (data - bufdata);
//      g_print("leaving %d bytes at %d for next chain()\n",
//              mpg123->lastframebytes,mpg123->lastframe);
//      g_print("handled %d frames this buffer\n",frames);
      return;
    }

/*
    g_print("we have a frame @ +%d: MPEG v%s, layer %d, %d Kbps, %d bytes\n",
            data-bufdata,
            (mpg123->frame->mpegver == MPG_VER_1) ? "1.0" :
            (mpg123->frame->mpegver == MPG_VER_2) ? "2.0" :
            (mpg123->frame->mpegver == MPG_VER_25) ? "2.5" : "UNK",
            (mpg123->frame->layer == MPG_LAYER_1) ? 1 :
            (mpg123->frame->layer == MPG_LAYER_2) ? 2 :
            (mpg123->frame->layer == MPG_LAYER_3) ? 3 : 0,
            mpg123->frame->bitrate,mpg123->frame->bpf);
*/

    /* skip the CRC if it exists */
    if (mpg123->frame->protection)
      data += 2;

    /* now we want to actually decode the frame */
    g_print("about to decode a frame\n");
    mpg123_decode_frame(mpg123->frame,
                        data,size-(data - bufdata),
                        outbuf,1152*4);

    data += mpg123->frame->bpf;
//    offset += mpg123->frame->bpf;
//    g_print("skipping %d bytes of decoded audio\n",mpg123->frame->bpf);

    /* we've handled yet another frame, whee! */
    frames++;
  }
  g_print("handled %d frames this buffer\n",frames);
  gst_buffer_unref(buf);
}


GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;
  guint16 mp3type, rawtype;

  if (gst_plugin_find("mpg123") != NULL) return NULL;

  plugin = gst_plugin_new("mpg123");
  g_return_if_fail(plugin != NULL);

  /* create an elementfactory for the mpg123 element */
  factory = gst_elementfactory_new("mpg123",GST_TYPE_MPG123,
                                   &gst_mpg123_details);
  g_return_if_fail(factory != NULL); 
  gst_plugin_add_factory(plugin,factory);
  g_print("mpg123: added 'mpg123' element\n");

  /* this filter inputs mp3 audio */
  mp3type = gst_type_find_by_mime("audio/mp3");
  g_return_if_fail(mp3type != 0);
  gst_type_add_sink(mp3type,factory);

  /* this filter outputs raw audio */
  rawtype = gst_type_find_by_mime("audio/raw");
  g_return_if_fail(rawtype != 0);
  gst_type_add_src(rawtype,factory);

  return plugin;
}
