/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef __GST_MPG123_H__
#define __GST_MPG123_H__


#include <config.h>
#include <gst/gst.h>
#include <gst/meta/audioraw.h>

#include "mpg123.h"
#include "mpglib.h"


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define GST_TYPE_MPG123 \
  (gst_mpg123_get_type())
#define GST_MPG123(obj) \
  (GTK_CHECK_CAST((obj),GST_TYPE_MPG123,GstMpg123))
#define GST_MPG123_CLASS(klass) \
  (GTK_CHECK_CLASS_CAST((klass),GST_TYPE_MPG123,GstMpg123))
#define GST_IS_MPG123(obj) \
  (GTK_CHECK_TYPE((obj),GST_TYPE_MPG123))
#define GST_IS_MPG123_CLASS(obj) \
  (GTK_CHECK_CLASS_TYPE((klass),GST_TYPE_MPG123)))

typedef struct _GstMpg123 GstMpg123;
typedef struct _GstMpg123Class GstMpg123Class;

struct _GstMpg123 {
  GstFilter filter;

  /* pads */
  GstPad *sinkpad,*srcpad;

  /* state */
  struct frame *frame;		/* mpg123 structure for frame */
  gulong prevheader;		/* previous header, to save time */

  /* in case we have something left over */
  GstBuffer *prevbuf;		/* the previous buffer */
  guint lastframe;		/* byte offset of last frame */
  guint lastframebytes;		/* bytes of last frame in prevbuf */
  guchar *framebuf;		/* memory region for reconstruction */
  gint framebufsize;		/* allocated size of region */

  /* error counters */
  gulong skipped_bytes;		/* how many bytes did we have to skip */
  gulong skipped_buffers;	/* how many buffers were useless */

  /* mpglib stuff */
  struct mpstr mp;
};

struct _GstMpg123Class {
  GstFilterClass parent_class;
};

GtkType gst_mpg123_get_type(void);
GstElement *gst_mpg123_new(gchar *name);
void gst_mpg123_chain(GstPad *pad,GstBuffer *buf);


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __GST_MPG123_H__ */
