/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#define PCM_BUFFER_SIZE (1152*4)

#include <gstxing.h>


GstElementDetails gst_xing_details = {
  "Xing mp3 decoder",
  "Filter/Decoder/Audio",
  "Uses Xing codec to decode mp3 streams",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* Xing signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_xing_class_init(GstXingClass *klass);
static void gst_xing_init(GstXing *xing);


static GstFilterClass *parent_class = NULL;
static guint gst_xing_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_xing_get_type(void) {
  static GtkType xing_type = 0;

  if (!xing_type) {
    static const GtkTypeInfo xing_info = {
      "GstXing",
      sizeof(GstXing),
      sizeof(GstXingClass),
      (GtkClassInitFunc)gst_xing_class_init,
      (GtkObjectInitFunc)gst_xing_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    xing_type = gtk_type_unique(GST_TYPE_FILTER,&xing_info);
  }
  return xing_type;
}

static void
gst_xing_class_init(GstXingClass *klass) {
  GstFilterClass *gstfilter_class;

  gstfilter_class = (GstFilterClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_FILTER);
}

static void gst_xing_init(GstXing *xing) {
  /* create the sink and src pads */
  xing->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(xing),xing->sinkpad);
  gst_pad_set_chain_function(xing->sinkpad,gst_xing_chain);
  xing->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(xing),xing->srcpad);
}

GstElement *gst_xing_new(gchar *name) {
  GstElement *xing = GST_ELEMENT(gtk_type_new(GST_TYPE_XING));
  gst_object_set_name(GST_OBJECT(xing),name);
  return xing;
}

void gst_xing_chain(GstPad *pad,GstBuffer *buf) {
  GstXing *xing;
  /* buffer info */
  guchar *bufdata, *data;
  gulong size;
  /* reconstruction pointers */
  guchar *framedata;
  /* per-buffer state */
  gint frames = 0;
  gulong header;

  guchar *pcm_buffer;
  GstBuffer *newbuf;
  IN_OUT inout;

  g_return_if_fail(pad != NULL);
  g_return_if_fail(GST_IS_PAD(pad));
  g_return_if_fail(buf != NULL);
  g_return_if_fail(GST_IS_BUFFER(buf));

  xing = GST_XING(pad->parent);

  bufdata = (guchar *)buf->data;
  data = bufdata;
  size = buf->datasize;
//  g_print("gst_xing_chain: got buffer of %d bytes in '%s'\n",size,
//          gst_object_get_name(GST_OBJECT(xing)));

  /* first deal with an underflowed frame in previous buffer */
  /* xing->header should still be valid? */
  if (xing->prevbuf != NULL) {
    g_print("dealing with previous unfinished buffer\n");
    /* first verify we have a buffer of sufficient size to construct to */
    if (xing->framebuf == NULL) {
      g_print("allocating new buffer of %d bytes\n",xing->framebytes);
      xing->framebuf = (guchar *)malloc(xing->framebytes);
    } else if (xing->framebufsize < xing->framebytes) {
      g_print("allocating a bigger buffer of %d bytes\n",xing->framebytes);
      free(xing->framebuf);
      xing->framebuf = (guchar *)malloc(xing->framebytes);
    }

    /* now copy the first part of the frame over */
    memcpy(xing->framebuf,xing->prevbuf->data+xing->lastframe,
           xing->lastframebytes);
    /* and free the buffer */
    gst_buffer_destroy(xing->prevbuf);

    /* now put the second part in */
    memcpy(xing->framebuf + xing->framebytes,data,
           xing->framebytes - xing->lastframebytes);

    data += (xing->lastframebytes - xing->lastframebytes);

    inout = audio_decode(xing->framebuf,(guint16 *)pcm_buffer);
//    g_print("decoded %d bytes mp3 to %d bytes pcm\n",inout.in_bytes,
//            inout.out_bytes);

    if (inout.out_bytes > 0) {
      newbuf = GST_BUFFER(gst_buffer_new());
      g_return_if_fail(newbuf != NULL);
      newbuf->meta == NULL;
      newbuf->data = pcm_buffer;
      newbuf->datasize = inout.out_bytes;
      gst_pad_push(xing->srcpad,newbuf);
    }

    frames++;
  }

  /* loop until we're out of buffer */
  while ((data - bufdata) < size) {
    /* first seach for the beginning the possible beginning of a frame */
    while (*data != 0xff) {
      data++;
      xing->skipped_bytes++;
      /* if we're at the end of the buffer with no frames, squawk */
      if ((data - bufdata) == size) {
        xing->skipped_buffers++;
        /* FIXME: should be dereferencing the buffer, not destroying */
        gst_buffer_destroy(buf);
        /* FIXME: fire off a signal here */
        g_print("sorry, couldn't find a byte of sync\n");
        return;
      }
    }

//    g_print("checking for header at %d\n",(data - bufdata));

    /* snag the frame header for future use */
    header = GULONG_TO_BE(*((gulong *)data));
    /* now check the potential frame header */
    if (!(xing->framebytes = head_info2(data,size - (data - bufdata),
                                        &xing->header,&xing->bitrate)))
      break;

    /* fudge the framesize, since xing gets it wrong (?!?!?!?!) */
    xing->framebytes++;

//    g_print("we have a frame of %d bytes!\n",xing->framebytes);

    /* at this point we should have a valid MPEG audio frame header */

    /* initialize stuff if we haven't already decoded this header... */
    if (header != xing->prevheader) {
      if (!audio_decode_init(&xing->header,xing->framebytes,0,0,0,24000)) {
        g_print("sorry, decoder wouldn't initialize...\n");
        return;
      }
      g_print("new header, br_index is %d, sr_index is %d\n",
              xing->header.br_index,xing->header.sr_index);
      xing->prevheader = header;
    }

    /* deal with the case where we don't have enough data this buffer */
    if ((size - (data-bufdata)) < xing->framebytes) {
      g_print("frame underflow, need next buffer to complete frame\n");
      xing->prevbuf = buf;
      xing->lastframe = data - bufdata;
      xing->lastframebytes = size - (data - bufdata);
      g_print("leaving %d bytes at %d for next chain()\n",
              xing->lastframebytes,xing->lastframe);
      g_print("handled %d frames this buffer\n",frames);
      return;
    }

    /* now we want to actually decode the frame */
    pcm_buffer = (guchar *)malloc(PCM_BUFFER_SIZE);
    g_return_if_fail(pcm_buffer != NULL);

    inout = audio_decode(data,(guint16 *)pcm_buffer);
//    g_print("decoded %d bytes mp3 to %d bytes pcm\n",inout.in_bytes,
//            inout.out_bytes);

    if (inout.out_bytes > 0) {
      newbuf = GST_BUFFER(gst_buffer_new());
      g_return_if_fail(newbuf != NULL);
      newbuf->meta == NULL;
      newbuf->data = pcm_buffer;
      newbuf->datasize = inout.out_bytes;
      gst_pad_push(xing->srcpad,newbuf);
    } else
      g_free(pcm_buffer);

    data += inout.in_bytes;

    /* we've handled yet another frame, whee! */
    frames++;
  }
  g_print("handled %d frames this buffer\n",frames);
  gst_buffer_destroy(buf);
}


GstPlugin *plugin_init(GModule *module) {
  GstPlugin *plugin;
  GstElementFactory *factory;
  guint16 mp3type, rawtype;

  if (gst_plugin_find("xing") != NULL) return NULL;

  plugin = gst_plugin_new("xing");
  g_return_if_fail(plugin != NULL);

  /* create an elementfactory for the mp3 element */
  factory = gst_elementfactory_new("xing",GST_TYPE_XING,
                                   &gst_xing_details);
  g_return_if_fail(factory != NULL);
  gst_plugin_add_factory(plugin,factory);
  g_print("xing: added 'xing' element\n");
 
  /* this filter inputs mp3 audio */
  mp3type = gst_type_find_by_mime("audio/mp3");
  g_return_if_fail(mp3type != 0);
  gst_type_add_sink(mp3type,factory);

  /* this filter outputs raw audio */
  rawtype = gst_type_find_by_mime("audio/raw");
  g_return_if_fail(rawtype != 0);
  gst_type_add_src(rawtype,factory);
 
  return plugin;
}
